@@ -1,5 +1,7 @@
 package WWW::Babelfish;
 
+require 5.008;
+
 use strict;
 use vars qw($VERSION @ISA @EXPORT @EXPORT_OK);
 
@@ -12,7 +14,7 @@ require AutoLoader;
 # Do not simply export all your public functions/methods/constants.
 @EXPORT = qw();
 
-$VERSION = '0.11';
+$VERSION = '0.12';
 
 # Preloaded methods go here.
 
@@ -20,15 +22,75 @@ use HTTP::Request::Common qw(POST);
 use LWP::UserAgent;
 use HTML::TokeParser;
 use IO::String;
+use Encode;
 
-my $BABELFISH = 'babelfish.altavista.com';
-my $BABELFISH_URL = 'http://babelfish.altavista.com/babelfish/tr';
 my $MAXCHUNK = 1000; # Maximum number of characters 
-# Bablefish will translate at one time 
+                     # Bablefish will translate at one time 
 
 my $MAXRETRIES = 5;  # Maximum number of retries for a chunk of text
 $| = 1;
 
+my $Services = {
+		Babelfish => {
+			      agent => $0 . ":" . __PACKAGE__ . "/" . $VERSION,
+
+			      languagesrequest => sub {
+				my $req = new HTTP::Request(GET => 'http://babelfish.altavista.com/babelfish/tr?il=en');
+				return $req;
+			      },
+
+			      translaterequest => sub {
+				my($langpair, $text) = @_;
+				my $req = POST ( 'http://babelfish.altavista.com/babelfish/tr?il=en',
+						[ 'doit' => 'done', 'urltext' => encode("utf8",$text), 'lp' => $langpair, 'Submit' => 'Translate', 'enc' => 'utf8' ]);
+				return $req;
+			      },
+
+			      # Extract the text from the html we get back from babelfish and return
+			      # it (keying on the fact that it's the first thing after a <br> tag,
+			      # possibly removing a textarea tag after it).
+			      extract_text => sub {
+				my($html) = @_;
+				my $p = HTML::TokeParser->new(\$html);
+				my $tag;
+				while ($tag = $p->get_tag('input')) {
+				  $_ = @{$tag}[1]->{value} if @{$tag}[1]->{name} eq 'q';
+				  return decode("utf8",$_);
+				}
+			      }
+			     },
+
+		Google =>    {
+			      agent => 'Mozilla/5.0',  # Google is finicky
+
+			      languagesrequest => sub {
+				my $req = new HTTP::Request(GET => 'http://www.google.com/language_tools?hl=en');
+				return $req;
+			      },
+
+			      translaterequest => sub {
+				my($langpair, $text) = @_;
+				my $req = POST ( 'http://translate.google.com/translate_t',
+					     [ 'text' => encode("utf8",$text), 'langpair' => $langpair, hl => 'en', ie => "UTF8", oe => "UTF8",]);
+				return $req;
+			      },
+
+			      extract_text => sub {
+				my($html) = @_;
+				my $p = HTML::TokeParser->new(\$html);
+				my $tag;
+				while ($tag = $p->get_tag('textarea')) {
+				  if(@{$tag}[1]->{name} eq 'q'){
+				    $_ = $p->get_text;
+				    return decode("utf8",$_);
+				  }
+				}
+			      }
+			     }
+				
+	       };
+
+
 sub new {
     my ($this, @args) = @_;
     my $class = ref($this) || $this;
@@ -39,18 +101,23 @@ sub new {
 }
 
 sub initialize {
-    my($self, %params) = @_;;
+    my($self, %params) = @_;
+
+    $self->{service} = $params{service} || 'Babelfish';
+    die "No such service: " . $self->{service} unless defined $Services->{ $self->{service} };
 
     # Caller can set user agent; we default to "script:WWW::Babelfish/0.01"
-    $self->{agent} = $params{agent} || $0 . ":" . __PACKAGE__ . "/" . $VERSION;
+    $self->{agent} = $params{agent} || $Services->{agent};
 
     $self->{proxy} = $params{proxy} if defined $params{proxy};
 
     # Get the page 
     my $ua = new LWP::UserAgent;
     $ua->proxy('http','http://' . $self->{proxy}) if defined $self->{proxy};
-    $ua->agent($self->{agent});     
-    my $req = new HTTP::Request('GET' => $BABELFISH_URL);
+    $ua->agent($self->{agent});
+    $self->{ua} = $ua;
+
+    my $req = &{ $Services->{ $self->{service} }->{languagesrequest} };
     my $res = $ua->request($req);
     unless($res->is_success){ 
 	warn(__PACKAGE__ . ":" . $res->status_line);
@@ -58,18 +125,17 @@ sub initialize {
     }
     my $page = $res->content;
 
-    # Extract the mapping of languages to options to be passed back,
-    # and store it on our object in "Lang2opt" hash
-    # and extract language names and store on "Langs" hash 
+    # Extract the language names and the mapping of languages to options to
+    # be passed back, and store them on our object in "Langs" hash of hashes
+    # Incredibly, this works for both Babelfish and Google; it should really 
+    # be a method in $Services
     my $p = HTML::TokeParser->new(\$page);
     my $a2b;
     if( $p->get_tag("select") ){
 	while( $_ = $p->get_tag("option") ){
-	    $a2b = $p->get_trimmed_text;
-	    $self->{Lang2opt}->{$a2b} = $_->[1]{value};
-	    $a2b =~ /(\S+)\sto\s(\S+)/;
-	    $self->{Langs}->{$1} = ""; 
-	    $self->{Langs}->{$2} = ""; 
+	    ($a2b = $p->get_trimmed_text) =~ /(\S+)\sto\s(\S+)/ or next;
+	    $self->{Langs}{$1}{$2} = $_->[1]{value};
+            $self->{Langs}{$2} ||= {};
 	}
     }
 
@@ -81,49 +147,53 @@ sub languages {
     return sort keys %{$self->{Langs}};
 }
 
+sub languagepairs {
+    my $self = shift;
+    return $self->{Langs};
+}
+
 sub translate {
   my ($self, %params) = @_;
-  
+
   # Paragraph separator is "\n\n" by default
   $/ = $params{delimiter} || "\n\n";
-  
+
   undef $self->{error};
   unless( exists($self->{Langs}->{$params{source}}) ){
-    $_ = "Language \"" . $params{source} . "\" is not available";
-    $self->{error} = $_;
-    warn(__PACKAGE__ . ": " . $_);
+    $self->{error} = qq(Language "$params{source}" is not available);
+    warn(__PACKAGE__ . ": " . $self->{error} . "\n");
     return undef;
   }
-  
-  unless( exists($self->{Langs}->{$params{destination}}) ){
-    $_ =  "Language \"" . $params{destination} . "\" is not available";
-    $self->{error} = $_;
-    warn(__PACKAGE__ . ": " . $_);
-    return undef;
-  }
-  
+
   # This "feature" is actually useful as a pass-thru filter.
   # Babelfish doesn't do same-to-same anyway (though it would be
   # pretty interesting if it did)
   return $params{text} if $params{source} eq $params{destination};
-  
-  my $langopt = $self->{Lang2opt}->{$params{source} . " to " . $params{destination}};
-  
+
+  unless ( exists($self->{Langs}->{$params{source}}{$params{destination}}) ){
+    $self->{error} =
+          qq(Cannot translate from "$params{source}" to "$params{destination}");
+    warn(__PACKAGE__ . ": " . $self->{error} . "\n");
+    return undef;
+  }
+
+  my $langopt = $self->{Langs}{$params{source}}{$params{destination}};
+
   my $th;			# "Text Handle"
   if( ref $params{text} ){	# We've been passed a filehandle
-    $th = $params{text};  
+    $th = $params{text};
   }
   else{				# We've been passed a string
     $th = new IO::String($params{text});
   }
-  
+
   my $Text = "";
   my $WANT_STRING_RETURNED = 0;
   unless( defined $params{ofh} ){
     $params{ofh} = new IO::String($Text);
     $WANT_STRING_RETURNED = 1;
   }
-  
+
   # Variables we use in the next mega-block
   my $para;			# paragraph
   my $num_paras = 0;		# number of paragraphs
@@ -138,36 +208,32 @@ sub translate {
   while($para = <$th>){
     $num_paras++;
     $transpara = "";
-    
+
     # Extract any leading whitespace from the start of the paragraph
     # Babelfish will eat it anyway.
     if ($para =~ s/(^\s+)(\S)/$2/) {
 	$para_start_ws = $1 || "";
     }
     chomp $para;		# Remove the para delimiter
-    
+
   CHUNK:
     foreach $chunk ( $self->_chunk_text($MAXCHUNK, $para) ) {
-      $req = POST ($BABELFISH_URL, [ 'doit' => 'done', 'urltext' => $chunk, 'lp' => $langopt, 'Submit' => 'Translate', 'enc' => 'utf8' ]);
-      $ua = new LWP::UserAgent;
-      $ua->proxy('http','http://' . $self->{proxy}) if defined $self->{proxy};
-      
+      $req = &{ $Services->{ $self->{service} }->{translaterequest} }($langopt, $chunk);
+      $ua = $self->{ua};
+
     RETRY:
       for($i = 0; $i <= $MAXRETRIES; $i++){ 
 	$res = $ua->request($req);
+
 	if( $res->is_success ){
 
-	  #DEBUG
-	  open(DEBUG, ">debug.html");
-	  print DEBUG $res->as_string;
-	  close DEBUG;
-
-	  $text = $self->_extract_text($res->as_string);
+	  #$text = $self->_extract_text($res->as_string); #REMOVE
+	  $text = &{ $Services->{ $self->{service} }->{extract_text} }($res->as_string);
 	  next RETRY if $text =~ /^\*\*time-out\*\*/; # in-band signalling; yuck
-	  
+
 	  $text =~ s/\n$//;	# Babelfish likes to append newlines
 	  $transpara .= $text;
-	  
+
 	  next CHUNK;
 	}
       }
@@ -177,8 +243,7 @@ sub translate {
     print { $params{ofh} } $/ if $num_paras > 1;
     print { $params{ofh} } $para_start_ws . $transpara;
   }
-  
-  close $params{ofh};
+
   if( $WANT_STRING_RETURNED ){
     return $Text;
   }
@@ -192,14 +257,14 @@ sub error {
   return $self->{error};
 }
 
-# Given a maximum chunk size and some text, return 
+# Given a maximum chunk size and some text, return
 # an array of pieces of the text chopped up in a 
 # logical way and less than or equal to the chunk size
 sub _chunk_text {
     my($self, $max, $text) = @_;
-    
+
     my @result;
-    
+
     # The trivial case
     return($text) if length($text) <= $max; 
 
@@ -207,6 +272,11 @@ sub _chunk_text {
     # I'm guessing that Babelfish doesn't look at any structure larger than 
     # a sentence; in fact I'm often tempted to guess that it doesn't look
     # at anything larger than a word, but we'll give it the benefit of the doubt.
+    #
+
+    # FIXME there are no built-in regexps for matching sentence
+    # breaks; I'm not sure if terminal punctuation will work for all
+    # languages...
 
     my $total = length($text);
     my $offset = 0;
@@ -237,7 +307,7 @@ sub _chunk_text {
 	    if( $offset == -1 or $offset <= $lastoffset){
 		$offset = rindex($text, " ", $test);
 	    }
-	    
+
 	    # or give up
 	    return undef if $offset == -1;
 	}
@@ -252,19 +322,6 @@ sub _chunk_text {
     return @result;
 }
 
-# Extract the text from the html we get back from babelfish and return
-# it (keying on the fact that it's the first thing after a <br> tag,
-# possibly removing a textarea tag after it).
-sub _extract_text {
-  my($self, $html) = @_;
-
-  my $p = HTML::TokeParser->new(\$html);
-
-  my $tag;
-  while ($tag = $p->get_tag('input')) {
-    return @{$tag}[1]->{value} if @{$tag}[1]->{name} eq 'q';
-  }
-
 # This code is now obsoleted by the new result page format, but I'm
 # leaving it here commented out in case we end up needing the
 # whitespace hack again.
@@ -289,7 +346,7 @@ sub _extract_text {
 #    return $text;
 
 
-}
+#}
 
 
 # Autoload methods go after =cut, and are processed by the autosplit program.
@@ -304,7 +361,7 @@ WWW::Babelfish - Perl extension for translation via babelfish
 =head1 SYNOPSIS
 
   use WWW::Babelfish;
-  $obj = new WWW::Babelfish( 'agent' => 'Mozilla/8.0', 'proxy' => 'myproxy' );
+  $obj = new WWW::Babelfish( service => 'Babelfish', agent => 'Mozilla/8.0', proxy => 'myproxy' );
   die( "Babelfish server unavailable\n" ) unless defined($obj);
 
   $french_text = $obj->translate( 'source' => 'English',
@@ -326,13 +383,74 @@ Perl interface to the WWW babelfish translation server.
 
 =item new
 
-Creates a new WWW::Babelfish object. It can take a named argument for
-user agent.
+Creates a new WWW::Babelfish object.
+
+Parameters:
+
+ service:        Babelfish or Google; default is Babelfish
+ agent:          user agent string
+ proxy:          proxy in the form of host:port
 
 =item languages
 
 Returns a plain array of the languages available for translation.
 
+=item languagepairs
+
+Returns a reference to a hash of hashes.
+The keys of the outer hash reflect all available languages.
+The hashes the corresponding values reference contain one (key) entry
+for each destination language that the particular source language can
+be translated to.
+The values of these inner hashes contain the Babelfish option name for
+the language pair.
+You should not modify the returned structure unless you really know
+what you're doing.
+
+Here's an example of a possible return value:
+
+	{
+	  'Chinese' => {
+	                 'English' => 'zh_en'
+	               },
+	  'English' => {
+	                 'Chinese' => 'en_zh',
+	                 'French' => 'en_fr',
+	                 'German' => 'en_de',
+	                 'Italian' => 'en_it',
+	                 'Japanese' => 'en_ja',
+	                 'Korean' => 'en_ko',
+	                 'Portuguese' => 'en_pt',
+	                 'Spanish' => 'en_es'
+	               },
+	  'French' => {
+	                'English' => 'fr_en',
+	                'German' => 'fr_de'
+	              },
+	  'German' => {
+	                'English' => 'de_en',
+	                'French' => 'de_fr'
+	              },
+	  'Italian' => {
+	                 'English' => 'it_en'
+	               },
+	  'Japanese' => {
+	                  'English' => 'ja_en'
+	                },
+	  'Korean' => {
+	                'English' => 'ko_en'
+	              },
+	  'Portuguese' => {
+	                    'English' => 'pt_en'
+	                  },
+	  'Russian' => {
+	                 'English' => 'ru_en'
+	               },
+	  'Spanish' => {
+	                 'English' => 'es_en'
+	               }
+	};
+
 =item translate
 
 Translates some text using Babelfish.
@@ -364,7 +482,13 @@ Returns a (hopefully) meaningful error string.
 Babelfish translates 1000 characters at a time. This module tries to
 break the source text into reasonable logical chunks of less than 1000
 characters, feeds them to Babelfish and then reassembles
-them. Formatting may get lost in the process.
+them. Formatting may get lost in the process; also it's doubtful this
+will work for non-Western languages since it tries to key on
+punctuation. What would make this work is if perl had properly
+localized regexps for sentence/clause boundaries.
+
+Support for Google is preliminary and hasn't been extensively tested.
+Google's translations are suspiciously similar to Babelfish's...
 
 =head1 AUTHOR
 
